package javax.swing;
import java.awt.Component;

//import javax.accessibility.Accessible;

/**
 * Component that shows an inactive text and/or icon.
 *
 * <p>The following layout rules apply:
 *
 * <p>If horizontalTextPosition is CENTER, the icon and the text is
 * stacked vertical, if horizontalTextPosition is anything other than
 * CENTER, the icon and text is stacked horizontally.
 *
 * <p>If stacking is vertical, the verticalTextPosition determines whether
 * the text appears above or below the icon.
 *
 * <p>If stacking is horizontal, the verticalTextPosition determines which
 * of the top, center or bottom of the bounding boxes of the icon and
 * text are vertically aligned.
 *
 *<pre>
 *
 *      hTextAlign     LEFT          CENTER           RIGHT
 * vTextAlign
 *                  +----+text        text          text+----+
 * TOP              |    |           +----+             |    |
 *                  +----+           |    |             +----+
 *                                   +----+
 *
 *                  +----+           +----+             +----+
 * CENTER           |    |text       |text|         text|    |
 *                  +----+           +----+             +----+
 *
 *                  +----+           +----+             +----+
 * BOTTOM           |    |           |    |             |    |
 *                  +----+text       +----+         text+----+
 *                                    text 
 *</pre>
 *
 * This is based on observations made of the behavior of the Sun Swing
 * implementation. As always, this has not been properly documented by
 * Sun. Layout policy does not seem to vary among the different PLAF
 * implementations.
 **/

public class JLabel extends JComponent implements SwingConstants
    //, Accessible
{
    protected Component labelFor;



    String text;
    Icon icon;
    Icon disIcon;
    
    int iconTextGap = 4;

    int vAlign = CENTER;
    int hAlign = LEADING;

    int vTextPos = CENTER;
    int hTextPos = TRAILING;

    public JLabel(String text, Icon image, int horizontalAlignment) {
	this.text = text;
	this.hAlign = horizontalAlignment;
	icon = image;
	disIcon = image;   // do something with image?

	updateUI();
    }
    
    public JLabel(String text, int horizontalAlignment) {
	this(text, null, horizontalAlignment);
	
    }

    public JLabel(String text) {
	this(text, LEADING);
    }

    public JLabel(Icon image, int horizontalAlignment) {
	this(null, image, horizontalAlignment);
    }

    public JLabel(Icon image) {
	this(image, CENTER);
    }

    public JLabel() {
	this("");
	
    }

    public LabelUI getUI() {
	return (LabelUI) ui;
    }

    public void setUI(LabelUI ui) {
	super.setUI((ComponentUI) ui);
    }

    public void updateUI() {
	ComponentUI ui = UIManager.getUI(this);
	setUI((LabelUI) UIManager.getUI(this));
    }

    public String getUIClassID() {
	return "LabelUI";
    }

    public String getText() {
	return text;
    }

    public void setText(String text) {
	this.text = text;
    }

    public Icon getIcon() {
	return icon;
    }

    public void setIcon(Icon icon) {
	this.icon = icon;
    }
    
    public Icon getDisabledIcon() {
	return disIcon;
    }

    public void setDisabledIcon(Icon disabledIcon) {
	disIcon = disabledIcon;
    }

    //public void setDisplayedMnemonic(int keyCode);
    //public void setDisplayedMnemonic(char keyChar);

    //public int getDisplayedMnemonic();

    //protected int checkHorizontalKey(int key, String message);

    //protected int checkVerticalKey(int key, String message);

    public int getIconTextGap() {
	return iconTextGap;
    }
    
    public void setIconTextGap(int iconTextGap) {
	this.iconTextGap = iconTextGap;
    }
     
    /** 
	Returns the vertical aligment of the label contents. The
	default is SwingConstants.CENTER.

	@returns one of CENTER, TOP, BOTTOM from SwingContants. */
   public int getVerticalAlignment() {
	return vAlign;
    }

    /** @param vAlignment of CENTER, TOP, BOTTOM from SwingContants. */
    public void setVerticalAlignment(int alignment) {
	this.vAlign = alignment;
    }

    /**
       Returns the horizontal aligment of the label contents. The
       default is SwingConstants.LEADING for text only labels, and
       SwingConstants.CENTER if the label contains an Icon. */

    public int getHorizontalAlignment() {
	return hAlign;
    }

    public void setHorizontalAlignment(int alignment) {
	hAlign = alignment;
    }

    public int getVerticalTextPosition() {
	return vTextPos;
    }

    public void setVerticalTextPosition(int position) {
	vTextPos = position;
    }

    public int getHorizontalTextPosition() {
	return hTextPos;
    }

    public void setHorizontalTextPosition(int position) {
	hTextPos = position;
    }

    /*
    public boolean imageUpdate(Image img, int infoflags,
			       int x, int y, int width, int height) {
	
    }
    */

    protected String paramString() {
	return super.paramString() + ",text="+getText();
    }    

    public Component getLabelFor() {
	return labelFor;
    }

    public void setLabelFor(Component component) {
	labelFor = component;
    }

    /*
    public AccessibleContext getAccessibleContext() {
    }
    */

}
